#include <omp.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

// assume temperature is in celsius
#define MIN_TEMP 0 // ocean can't be frozen
#define MAX_TEMP 100 // ocean can't be boiling

void average(float** global_ocean, float** local_ocean, int x_dim, int y_dim) {
  int i, j;
  
  // the outer edge of the ocean doesn't participate
  #pragma omp parallel for \
    shared(x_dim, y_dim, local_ocean, global_ocean) \
    private(i,j) \
    schedule(dynamic)
  for (j = 1; j < y_dim - 1; j++) {
      for (i = 1; i < x_dim - 1; i++) {
      local_ocean[i][j] = (global_ocean[i][j] + global_ocean[i-1][j] + 
                           global_ocean[i][j+1] + global_ocean[i+1][j] + 
                           global_ocean[i][j-1]) / 5.0;
    }
  }
/*
  for (i = 1; i < x_dim - 1; i++) {
    for (j = 1; j < y_dim - 1; j++) {
      local_ocean[i][j] = (global_ocean[i][j] + global_ocean[i-1][j] + 
                           global_ocean[i][j+1] + global_ocean[i+1][j] + 
                           global_ocean[i][j-1]) / 5.0;
    }
  }
*/
}

void init_hand(float** ocean, int x_dim, int y_dim) {
  int i, j;

  for (i = 0; i < y_dim; i++) {
    ocean[0][i] = 10;
    ocean[x_dim - 1][i] = 0;
  }
  for (i = 1; i < x_dim - 1; i++) {
    ocean[i][0] = 10;
    ocean[i][y_dim - 1] = 0;
  }
}

void init_rand(float** ocean, int x_dim, int y_dim) {
  int i, j;
  int temp_range = MAX_TEMP - MIN_TEMP;

  srand(time(0));

  #pragma omp parallel for \
    shared(x_dim, y_dim, ocean) \
    private(i,j) \
    schedule(dynamic)
  for (i = 0; i < x_dim; i++) {
    for (j = 0; j < y_dim; j++) {
      ocean[i][j] = ((((float) rand() / RAND_MAX) * temp_range)) + MIN_TEMP;
    }
  }
}

void print_ocean(float** ocean, int x_dim, int y_dim) {
  int i, j;

  printf("*** OCEAN ***\n");
  for (i = 0; i < x_dim; i++) {
    for (j = 0; j < y_dim; j++) {
      printf("%d\t", (int) ocean[i][j]);
    }
    printf("\n");
  }
}

int main(int argc, char** argv) {

  int i, x_dim, y_dim, num_timesteps, num_threads;
  float **ocean_global, **ocean_local, **temp;
  hrtime_t start_time, end_time;
  
  // parse input
  if (argc < 5) {
    printf("**************************************************************\n");
    printf("* USAGE:                                                     *\n");
    printf("* ocean-dyn <x_dim> <y_dim> <num_timesteps> <num_threads>    *\n");
    printf("* <x_dim>         - the size of the grid's x-dimension       *\n");
    printf("* <y_dim>         - the size of the grid's y-dimension       *\n");
    printf("* <num_timesteps> - the number of updates to simulate        *\n");
    printf("* <num_threads>   - the number of threads                    *\n");
    printf("**************************************************************\n");
    return -1;
  } else {
    x_dim = atoi(argv[1]);
    y_dim = atoi(argv[2]);
    num_timesteps = atoi(argv[3]);
    num_threads = atoi(argv[4]);
    /*
    printf("x_dim: %d\n", x_dim);
    printf("y_dim: %d\n", y_dim);
    printf("num_timesteps: %d\n", num_timesteps);
    printf("num_threads: %d\n", num_threads);
    */
  }

  omp_set_num_threads(num_threads);

  // allocate oceans
  ocean_global = malloc(x_dim * sizeof(float*));
  ocean_local = malloc(x_dim * sizeof(float*));
  for (i = 0; i < x_dim; i++) {
    ocean_global[i] = malloc(y_dim * sizeof(float));
    ocean_local[i] = malloc(y_dim * sizeof(float));
  }

  // generate data
  init_rand(ocean_global, x_dim, y_dim);
  //init_hand(ocean_global, x_dim, y_dim);
  for (i = 0; i < y_dim; i++) {
    ocean_local[0][i] = ocean_global[0][i];
    ocean_local[x_dim - 1][i] = ocean_global[x_dim - 1][i];
  }
  for (i = 1; i < x_dim - 1; i++) {
    ocean_local[i][0] = ocean_global[i][0];
    ocean_local[i][y_dim - 1] = ocean_global[i][y_dim - 1];
  }

  /*
  printf("*** INIT ");
  print_ocean(ocean_global, x_dim, y_dim);
  */

  start_time = gethrtime();
  
  for (i = 0; i < num_timesteps; i++) {
    average(ocean_global, ocean_local, x_dim, y_dim);

    // swap
    temp = ocean_global;
    ocean_global = ocean_local;
    ocean_local = temp;
    /*
    if (i % (num_timesteps / 10) == 0) {
      printf("*** %d ", i);
      print_ocean(ocean_global, x_dim, y_dim);
    }
    */
  }

  end_time = gethrtime();
  printf("%f\n", (end_time - start_time) / 1000000000.0);

  /*
  printf("*** FINAL ");
  print_ocean(ocean_global, x_dim, y_dim);
  */

  return 0;
}
